import typing

import numpy
import torch
import sklearn.base
import sklearn.metrics
import torch.utils.data

import cvtda.utils
import cvtda.logging
import cvtda.neural_network


class NNClassifier(sklearn.base.ClassifierMixin):
    def __init__(
        self,

        n_jobs: int = -1,
        random_state: int = 42,

        device: torch.device = torch.device("cuda"),
        batch_size: int = 256,
        learning_rate: float = 1e-3,
        n_epochs: int = 100,
        
        skip_diagrams: bool = False,
        skip_images: bool = False,
        skip_features: bool = False
    ):
        self.n_jobs_ = n_jobs
        self.random_state_ = random_state

        self.device_ = device
        self.batch_size_ = batch_size
        self.learning_rate_ = learning_rate
        self.n_epochs_ = n_epochs

        self.skip_diagrams_ = skip_diagrams
        self.skip_images_ = skip_images
        self.skip_features_ = skip_features


    def fit(self, train: cvtda.neural_network.Dataset, val: typing.Optional[cvtda.neural_network.Dataset]):
        cvtda.utils.set_random_seed(self.random_state_)
        train_dl = torch.utils.data.DataLoader(
            torch.utils.data.TensorDataset(train.images, train.features, train.labels, torch.arange(len(train))),
            batch_size = self.batch_size_,
            shuffle = True
        )
        self.init_(len(torch.unique(train.labels)), *next(iter(train_dl)), train)

        for epoch in range(self.n_epochs_):
            sum_loss = 0

            self.model_list_.train()
            for images, features, labels, diagram_idxs in cvtda.logging.logger().pbar(train_dl, desc = f"Epoch {epoch}"):
                self.optimizer_.zero_grad()
                pred = self.forward_(images, features, diagram_idxs, train)
                loss = torch.nn.functional.cross_entropy(pred, labels.to(self.device_), reduction = 'mean')
                loss.backward()
                self.optimizer_.step()
                sum_loss += loss.item()
            postfix = { 'loss': sum_loss, 'lr': self.optimizer_.param_groups[0]['lr'] }
            self.scheduler_.step()

            if val is not None:
                val_proba = self.predict_proba_(val)
                val_pred = numpy.argmax(val_proba, axis = 1)
                postfix['val_acc'] = sklearn.metrics.accuracy_score(val.labels, val_pred)

            print(f"Epoch {epoch}:", postfix)

        self.fitted_ = True
        return self
    
    def predict_proba(self, dataset: cvtda.neural_network.Dataset) -> numpy.ndarray:
        assert self.fitted_ is True, 'fit() must be called before predict_proba()'
        cvtda.utils.set_random_seed(self.random_state_)
        return self.predict_proba_(dataset)


    def init_(
        self,
        num_classes: int,
        images: torch.Tensor,
        features: torch.Tensor,
        labels: torch.Tensor,
        diagram_idxs: torch.Tensor,
        dataset: cvtda.neural_network.Dataset
    ):
        diagrams = [] if self.skip_diagrams_ else dataset.get_diagrams(diagram_idxs)

        is_no_topology = (self.skip_diagrams_ and self.skip_features_ and not self.skip_images_)
        print('Topology: ', not is_no_topology)
        print('Images: ', images.shape)
        print('Features: ', features.shape)

        if is_no_topology:
            images_output = num_classes
        else:
            images_output = 1024
        
        self.nn_base_ = cvtda.neural_network.NNBase(
            num_diagrams = len(diagrams) // 2,
            skip_diagrams = self.skip_diagrams_,
            skip_images = self.skip_images_,
            skip_features = self.skip_features_,
            images_n_channels = images.shape[1],
            images_output = images_output
        ).to(self.device_).train()

        self.model_ = torch.nn.Sequential(
            torch.nn.Dropout(0.4), torch.nn.LazyLinear(256), torch.nn.BatchNorm1d(256), torch.nn.GELU(),
            torch.nn.Dropout(0.3), torch.nn.Linear(256, 128), torch.nn.BatchNorm1d(128), torch.nn.GELU(),
            torch.nn.Dropout(0.2), torch.nn.Linear(128, 64), torch.nn.BatchNorm1d(64), torch.nn.GELU(),
            torch.nn.Dropout(0.1), torch.nn.Linear(64, 32), torch.nn.BatchNorm1d(32), torch.nn.GELU(),
            torch.nn.Linear(32, num_classes)
        ).to(self.device_).train()

        if is_no_topology:
            self.model_ = torch.nn.Identity()

        self.model_list_ = torch.nn.ModuleList([ self.nn_base_, self.model_ ])

        self.optimizer_ = torch.optim.AdamW(
            params = self.model_list_.parameters(),
            lr = self.learning_rate_ * 100
        )
        
        def lr_scheduler_lambda(epoch):
            if epoch < self.n_epochs_ // 10:
                return 1
            if epoch < self.n_epochs_ // 4:
                return 0.1
            if epoch < self.n_epochs_ // 2:
                return 0.01
            if epoch < 3 * self.n_epochs_ // 4:
                return 0.001
            return 0.0001
        self.scheduler_ = torch.optim.lr_scheduler.LambdaLR(self.optimizer_, lr_scheduler_lambda)

        self.forward_(images, features, diagram_idxs, dataset)
        if not is_no_topology:
            cvtda.logging.logger().print(f'Input to LazyLinear: {self.model_[1].in_features}')
        cvtda.logging.logger().print(f'Parameters: {sum(p.numel() for p in self.model_list_.parameters())}')

    def forward_(
        self,
        images: torch.Tensor,
        features: torch.Tensor,
        diagram_idxs: torch.Tensor,
        dataset: cvtda.neural_network.Dataset
    ) -> torch.Tensor:
        images = images.to(self.device_)
        features = features.to(self.device_)
        if self.skip_diagrams_:
            return self.model_(self.nn_base_(images, features))
        return self.model_(self.nn_base_(images, features, *dataset.get_diagrams(diagram_idxs)))

    def predict_proba_(self, dataset: cvtda.neural_network.Dataset) -> numpy.ndarray:
        dl = torch.utils.data.DataLoader(
            torch.utils.data.TensorDataset(dataset.images, dataset.features, torch.arange(len(dataset))),
            batch_size = self.batch_size_,
            shuffle = False
        )

        y_pred_proba = [ ]
        self.model_list_.eval()
        with torch.no_grad():
            for images, features, diagram_idxs in dl:
                y_pred_proba.append(self.forward_(images, features, diagram_idxs, dataset))
        y_pred_proba = torch.vstack(y_pred_proba)
        return torch.nn.functional.softmax(y_pred_proba, dim = 1).cpu().numpy()



import abc
import typing

import numpy
import sklearn.base

import cvtda.utils
import cvtda.dumping
import cvtda.logging

from .. import utils
import cvtda.dumping

class Extractor(sklearn.base.TransformerMixin):
    def __init__(
        self,
        n_jobs: int = -1,
        reduced: bool = True,
        only_get_from_dump: bool = False,
        **kwargs
    ):
        self.n_jobs_ = n_jobs
        self.reduced_ = reduced
        self.only_get_from_dump_ = only_get_from_dump

        self.kwargs_ = kwargs
        self.kwargs_['n_jobs'] = n_jobs
        self.kwargs_['reduced'] = reduced
        self.kwargs_['only_get_from_dump'] = only_get_from_dump

        self.fitted_ = False
        self.fit_dimensions_ = None
    

    def final_dump_name_(self, dump_name: typing.Optional[str] = None):
        return self.features_dump_(dump_name)
    
    def features_dump_(self, dump_name: typing.Optional[str]):
        return cvtda.dumping.dump_name_concat(dump_name, "features")
    
    def force_numpy_(self):
        return True
    

    def fit(self, images: numpy.ndarray, dump_name: typing.Optional[str] = None):
        if self.only_get_from_dump_ and (len(images.shape) != 4):
            final_dump = self.final_dump_name_(dump_name)
            assert cvtda.dumping.dumper().has_dump(final_dump), f"There is no dump at {final_dump}"
        else:
            self.process_(images, do_fit = True, dump_name = dump_name)
        self.fitted_ = True
        return self
    
    def transform(self, images: numpy.ndarray, dump_name: typing.Optional[str] = None):
        assert self.fitted_ is True, 'fit() must be called before transform()'
        final_dump = self.final_dump_name_(dump_name)
        if (self.only_get_from_dump_ and (len(images.shape) != 4)) or cvtda.dumping.dumper().has_dump(final_dump):
            return cvtda.dumping.dumper().get_dump(final_dump)
        return self.process_(images, do_fit = False, dump_name = dump_name)
    
    def fit_transform(self, images: numpy.ndarray, dump_name: typing.Optional[str] = None):
        return self.fit(images, dump_name = dump_name).transform(images, dump_name = dump_name)


    def process_(self, images: numpy.ndarray, do_fit: bool, dump_name: typing.Optional[str] = None):
        if self.fit_dimensions_ is not None:
            assert self.fit_dimensions_ == images.shape[1:], \
                    f"The pipeline is fit for {self.fit_dimensions_}. Cannot use it with {images.shape}."
        
        if len(images.shape) == 4:
            assert images.shape[3] == 3, f'Images with {len(images.shape)} channels are not supported'
            cvtda.logging.logger().print("RGB images received. Transforming to grayscale.")

            rgb_dump = cvtda.dumping.dump_name_concat(dump_name, "rgb")
            gray_dump = cvtda.dumping.dump_name_concat(dump_name, "gray")
            red_dump = cvtda.dumping.dump_name_concat(dump_name, "red")
            green_dump = cvtda.dumping.dump_name_concat(dump_name, "green")
            blue_dump = cvtda.dumping.dump_name_concat(dump_name, "blue")
            saturation_dump = cvtda.dumping.dump_name_concat(dump_name, "saturation")
            value_dump = cvtda.dumping.dump_name_concat(dump_name, "value")

            if do_fit:
                self.gray_extractor_ = self.__class__(**self.kwargs_)
                self.red_extractor_ = self.__class__(**self.kwargs_)
                self.green_extractor_ = self.__class__(**self.kwargs_)
                self.blue_extractor_ = self.__class__(**self.kwargs_)
                if not self.reduced_:
                    self.saturation_extractor_ = self.__class__(**self.kwargs_)
                    self.value_extractor_ = self.__class__(**self.kwargs_)

            if self.only_get_from_dump_:
                rgb_data = cvtda.dumping.dumper().get_dump(self.final_dump_name_(rgb_dump))
            else:
                rgb_data = self.process_rgb_(images, do_fit, rgb_dump)

            gray_images = cvtda.utils.rgb2gray(images, self.n_jobs_)
            result = [
                rgb_data,
                utils.process_iter(self.gray_extractor_, gray_images, do_fit, gray_dump),
                utils.process_iter(self.red_extractor_, images[:, :, :, 0], do_fit, red_dump),
                utils.process_iter(self.green_extractor_, images[:, :, :, 1], do_fit, green_dump),
                utils.process_iter(self.blue_extractor_, images[:, :, :, 2], do_fit, blue_dump),
            ]
            if not self.reduced_:
                hsv = cvtda.utils.rgb2hsv(images, self.n_jobs_)
                result.append(utils.process_iter(self.saturation_extractor_, hsv[:, :, :, 1], do_fit, saturation_dump))
                result.append(utils.process_iter(self.value_extractor_, hsv[:, :, :, 2], do_fit, value_dump))
            
            result = utils.hstack(result, self.force_numpy_())
        else:
            assert len(images.shape) == 3, f'{len(images.shape) - 1}d images are not supported'
            result = self.process_gray_(images, do_fit, dump_name)
    
        self.fit_dimensions_ = images.shape[1:]
        return result
    
    @abc.abstractmethod
    def process_rgb_(self, rgb_images: numpy.ndarray, do_fit: bool, dump_name: typing.Optional[str] = None):
        pass
    
    @abc.abstractmethod
    def process_gray_(self, gray_images: numpy.ndarray, do_fit: bool, dump_name: typing.Optional[str] = None):
        pass



import os
import typing

import numpy
import torch
import matplotlib.pyplot as plt

import cvtda.dumping
import cvtda.neural_network
from .BaseLearner import BaseLearner
from .DiagramsLearner import DiagramsLearner
from .NNLearner import NNLearner
from .SimpleTopologicalLearner import SimpleTopologicalLearner

def learn(
    train_images: numpy.ndarray,
    train_features: numpy.ndarray,
    train_labels: numpy.ndarray,
    train_diagrams: typing.List[numpy.ndarray],

    test_images: numpy.ndarray,
    test_features: numpy.ndarray,
    test_labels: numpy.ndarray,
    test_diagrams: typing.List[numpy.ndarray],

    n_jobs: int = 1,
    random_state: int = 42,
    dump_name: typing.Optional[str] = None,

    nn_device: torch.device = torch.device('cuda'),
    nn_batch_size: int = 64,
    nn_learning_rate: float = 1e-3,
    nn_epochs: int = 100,
    nn_margin: int = 0.1,
    nn_latent_dim: int = 256,
):
    nn_train = cvtda.neural_network.Dataset(
        train_images, train_diagrams, train_features, train_labels, n_jobs = n_jobs, device = nn_device
    )
    nn_test = cvtda.neural_network.Dataset(
        test_images, test_diagrams, test_features, test_labels, n_jobs = n_jobs, device = nn_device
    )

    def classify_one(learner: BaseLearner, name: str, display_name: str, ax: plt.Axes):
        print(f'Trying {name} - {learner}')
        learner.fit(nn_train, nn_test)
        ax.set_title(display_name)
        learner.estimate_quality(nn_test, ax)

    nn_kwargs = dict(
        n_jobs = n_jobs,
        random_state = random_state,
        device = nn_device,
        batch_size = nn_batch_size,
        learning_rate = nn_learning_rate,
        margin = nn_margin,
        latent_dim = nn_latent_dim
    )
    classifiers = [
        SimpleTopologicalLearner(n_jobs = n_jobs),
        DiagramsLearner(n_jobs = n_jobs),
        NNLearner(**nn_kwargs, n_epochs = nn_epochs,      skip_diagrams = True,  skip_images = False, skip_features = True),
        NNLearner(**nn_kwargs, n_epochs = nn_epochs * 2,  skip_diagrams = True,  skip_images = True,  skip_features = False),
        NNLearner(**nn_kwargs, n_epochs = nn_epochs,      skip_diagrams = True,  skip_images = False, skip_features = False),
        NNLearner(**nn_kwargs, n_epochs = nn_epochs // 4, skip_diagrams = False, skip_images = True,  skip_features = True)
    ]

    names = [
        'SimpleTopologicalLearner',
        'DiagramsLearner',
        'NNLearner_images',
        'NNLearner_features',
        'NNLearner_features_images',
        'NNLearner_diagrams'
    ]
    display_names = [
        'Топологические признаки',
        'Диаграммы устойчивости',
        'ResNet50 – базовая модель',
        'Нейронная сеть для тополог. признаков',
        'Комбинированная нейронная сеть',
        'Обучаемая векторизация диаграмм'
    ]

    figure, axes = plt.subplots(2, 3, figsize = (12, 5))
    for args in zip(classifiers, names, display_names, axes.flat):
        classify_one(*args)

    handles, labels = axes.flat[0].get_legend_handles_labels()
    figure.legend(handles, labels, loc = (0.367, 0.83))

    figure.tight_layout()

    dumper = cvtda.dumping.dumper()
    if (dump_name is not None) and isinstance(dumper, cvtda.dumping.NumpyDumper):
        file = dumper.get_file_name_(cvtda.dumping.dump_name_concat(dump_name, "distributions"))
        os.makedirs(os.path.dirname(file), exist_ok = True)
        figure.savefig(file[:-4] + ".svg")
        figure.savefig(file[:-4] + ".png")



import torch
import torchph.nn.slayer
import torchvision.models


class NNBase(torch.nn.Module):
    def __init__(
        self,

        num_diagrams: int,
        skip_diagrams: bool = False,
        features_per_diagram: int = 64,

        skip_images: bool = False,
        images_n_channels: int = 3,
        images_output: int = 1024,

        skip_features: bool = False,
        features_output: int = 1024
    ):
        super().__init__()
        
        def make_slayer():
            return torchph.nn.slayer.SLayerExponential(features_per_diagram)
        self.slayers_ = None if skip_diagrams else torch.nn.ModuleList([ make_slayer() for _ in range(num_diagrams) ])
        self.slayers_compressor_ = torch.nn.Linear(num_diagrams * features_per_diagram, features_output)

        self.images_ = None
        if not skip_images:
            self.images_ = torchvision.models.resnet50(num_classes = images_output)
            self.images_.conv1 = torch.nn.Conv2d(images_n_channels, 64, kernel_size = 7, stride = 2, padding = 3, bias = False)

        self.features_ = None if skip_features else torch.nn.LazyLinear(features_output)
    
    def forward(self, images, features, *diagrams):
        result = [ ]
        
        if self.images_ is not None:
            result.append(self.images_(images))

        if self.slayers_ is not None:
            result.append(
                self.slayers_compressor_(
                    torch.cat([
                        self.slayers_[i]((diagrams[2 * i], diagrams[2 * i + 1], diagrams[2 * i].shape[1], len(images)))
                        for i in range(len(self.slayers_))
                    ], dim = 1)
                )
            )
    
        if self.features_ is not None:
            result.append(self.features_(features))

        return torch.cat(result, dim = 1)



from .base import BaseLogger
from .cli import CLILogger
from .devnull import DevNullLogger

def logger() -> BaseLogger:
    if BaseLogger.current_logger is None:
        BaseLogger.current_logger = CLILogger()
    return BaseLogger.current_logger



import typing

import numpy
import sklearn.base

import cvtda.dumping
from .process_iter import process_iter

def process_iter_dump(
    transformer: sklearn.base.TransformerMixin,
    data: numpy.ndarray,
    do_fit: bool,
    dump_name: typing.Optional[str] = None,
    *args,
    **kwargs
):
    if dump_name is None:
        return process_iter(transformer, data, do_fit, *args, **kwargs)
    
    if do_fit:
        transformer.fit(data, *args, **kwargs)
    return cvtda.dumping.dumper().execute(lambda: transformer.transform(data, *args, **kwargs), dump_name)



import typing

import tqdm
import tqdm.contrib

from .base import BaseLogger

T = typing.TypeVar("T")

class CLILogger(BaseLogger):
    def __init__(self):
        pass

    def print(self, data: T) -> None:
        print(data)

    def pbar(
        self,
        data: typing.Iterable[T],
        total: int = None,
        desc: typing.Optional[str] = None
    ) -> typing.Iterable[T]:
        return tqdm.tqdm(data, total = total, desc = desc)

    def zip(
        self,
        *iterables, 
        desc: typing.Optional[str] = None
    ):
        return tqdm.contrib.tzip(*iterables, desc = desc)

    def set_pbar_postfix(self, pbar: tqdm.tqdm, data: dict):
        pbar.set_postfix(data)



import typing

import numpy
import gudhi.hera

import cvtda.neural_network
from .BaseLearner import BaseLearner


class DiagramsLearner(BaseLearner):
    def fit(self, train: cvtda.neural_network.Dataset, val: typing.Optional[cvtda.neural_network.Dataset]):
        pass

    def prepare_diagram_(self, diagram: numpy.ndarray, dimension: int) -> numpy.ndarray:
        dim_filter = (diagram[:, 2] == dimension)
        non_degenerate_filter = (diagram[:, 0] < diagram[:, 1])
        return diagram[dim_filter & non_degenerate_filter][:, 0:2]

    def calculate_distance_(self, first: int, second: int, dataset: cvtda.neural_network.Dataset):
        diagrams1 = dataset.raw_diagrams[first]
        diagrams2 = dataset.raw_diagrams[second]
        assert len(diagrams1) == len(diagrams2)

        distance_vector = []
        for i in range(0, len(diagrams1), 2):
            diag1 = diagrams1[i]
            diag2 = diagrams2[i]
            for dim in numpy.unique(diag1[:, 2]):
                d1 = self.prepare_diagram_(diag1, dim)
                d2 = self.prepare_diagram_(diag2, dim)
                distance_vector.append(gudhi.hera.bottleneck_distance(d1, d2))

        return numpy.sqrt(numpy.sum(numpy.array(distance_vector) ** 2))



import typing

import numpy
import joblib
import sklearn.base
import skimage.measure
import skimage.feature

import cvtda.utils
import cvtda.dumping
import cvtda.logging

from . import utils
from .interface import Extractor


def calc_curvature(gray_image: numpy.ndarray) -> numpy.ndarray:
    min_, max_= gray_image.min(), gray_image.max()
    assert (min_ >= 0) and (max_ <= 1), f'Bad image format: should be [0, 1]; received [{min_}, {max_}]'

    euler_numbers, area, perimeter = [], [], []
    for threshold in range(256):
        bin = (gray_image > threshold / 255.)
        euler_numbers.append(skimage.measure.euler_number(bin))
        area.append(bin.sum())
        perimeter.append(skimage.measure.perimeter(bin))
    series = numpy.array([ euler_numbers, area, perimeter ])
    series_diff = numpy.array([ numpy.diff(euler_numbers), numpy.diff(area), numpy.diff(perimeter) ])
    
    return numpy.concatenate([
        cvtda.utils.sequence2features(series).flatten(),
        cvtda.utils.sequence2features(series_diff).flatten(),
    ])

class GrayGeometryExtractor(sklearn.base.TransformerMixin):
    def __init__(self, n_jobs: int = -1, reduced: bool = True):
        self.fitted_ = False
        self.n_jobs_ = n_jobs
        self.reduced_ = reduced


    def fit(self, gray_images: numpy.ndarray):
        self.fitted_ = True
        return self

    def transform(self, gray_images: numpy.ndarray) -> numpy.ndarray:
        assert self.fitted_ is True, 'fit() must be called before transform()'
        
        def process_one_(gray_image: numpy.ndarray) -> numpy.ndarray:
            sift = skimage.feature.SIFT()
            try:
                sift.detect_and_extract(gray_image)
                sift_descriptors = sift.descriptors.transpose()
                if sift_descriptors.shape[1] == 0:
                    raise 'How is this possible?'
            except:
                sift_descriptors = numpy.zeros((128,1))
                
            orb = skimage.feature.ORB()
            try:
                orb.detect_and_extract(gray_image)
                orb_descriptors = orb.descriptors.transpose()
                if orb_descriptors.shape[1] == 0:
                    raise 'How is this possible?'
            except:
                orb_descriptors = numpy.zeros((256,1))

            basic_features = skimage.feature.multiscale_basic_features(gray_image).reshape((-1, 24))

            moments_central = skimage.measure.moments_central(gray_image, order = 9)
            moments_normalized = skimage.measure.moments_normalized(moments_central)

            image_shape = max(*gray_image.shape)
            daisy_parameters = dict(
                step = (6 * image_shape // 32),
                radius = (12 * image_shape // 32),
                rings = 5,
                histograms = 5,
                orientations = 8
            )

            try:
                inertia_tensor_eigvals = skimage.measure.inertia_tensor_eigvals(gray_image)
            except:
                inertia_tensor_eigvals = numpy.zeros((2))

            return numpy.nan_to_num(numpy.concatenate([
                skimage.feature.daisy(gray_image, **daisy_parameters).flatten(),
                cvtda.utils.sequence2features(numpy.ma.array(sift_descriptors), reduced = self.reduced_).flatten(),
                cvtda.utils.sequence2features(numpy.ma.array(orb_descriptors), reduced = self.reduced_).flatten(),
                skimage.feature.hog(gray_image, pixels_per_cell = (gray_image.shape[0] // 4, gray_image.shape[1] // 4)),
                cvtda.utils.sequence2features(numpy.ma.array(basic_features.transpose()), reduced = self.reduced_).flatten(),
                [ skimage.measure.blur_effect(gray_image) ],
                skimage.measure.centroid(gray_image),
                inertia_tensor_eigvals,
                skimage.measure.moments(gray_image, order = 9).flatten(),
                moments_central.flatten(),
                skimage.measure.moments_hu(moments_normalized).flatten(),
                [ skimage.measure.shannon_entropy(gray_image) ],
                calc_curvature(gray_image)
            ]), 0)

        return numpy.stack(
            joblib.Parallel(n_jobs = self.n_jobs_)(
                joblib.delayed(process_one_)(img)
                for img in cvtda.logging.logger().pbar(gray_images, desc = 'GrayGeometryExtractor')
            )
        )

class RGBGeometryExtractor(sklearn.base.TransformerMixin):
    def __init__(self, n_jobs: int = -1, reduced: bool = True):
        self.fitted_ = False
        self.n_jobs_ = n_jobs
        self.reduced_ = reduced


    def fit(self, rgb_images: numpy.ndarray):
        self.fitted_ = True
        return self

    def transform(self, rgb_images: numpy.ndarray) -> numpy.ndarray:
        assert self.fitted_ is True, 'fit() must be called before transform()'

        def process_one_(rgb_image: numpy.ndarray) -> numpy.ndarray:
            try:
                inertia_tensor_eigvals = skimage.measure.inertia_tensor_eigvals(rgb_image)
            except:
                inertia_tensor_eigvals = numpy.zeros((3))
            
            return numpy.nan_to_num(numpy.concatenate([
                skimage.feature.hog(rgb_image, channel_axis = 2),
                skimage.measure.centroid(rgb_image),
                inertia_tensor_eigvals,
                skimage.measure.moments(rgb_image, order = 6).flatten(),
                skimage.measure.moments_central(rgb_image, order = 6).flatten(),
                [
                    skimage.measure.pearson_corr_coeff(rgb_image[:, :, 0], rgb_image[:, :, 1])[0],
                    skimage.measure.pearson_corr_coeff(rgb_image[:, :, 0], rgb_image[:, :, 2])[0],
                    skimage.measure.pearson_corr_coeff(rgb_image[:, :, 1], rgb_image[:, :, 2])[0],
                ]
            ]), 0)

        return numpy.stack(
            joblib.Parallel(n_jobs = self.n_jobs_)(
                joblib.delayed(process_one_)(img)
                for img in cvtda.logging.logger().pbar(rgb_images, desc = 'RGBGeometryExtractor')
            )
        )


class GeometryExtractor(Extractor):
    def __init__(self, n_jobs: int = -1, reduced: bool = True, only_get_from_dump: bool = False):
        super().__init__(n_jobs = n_jobs, reduced = reduced, only_get_from_dump = only_get_from_dump)

        self.rgb_extractor_ = RGBGeometryExtractor(n_jobs = self.n_jobs_, reduced = self.reduced_)
        self.gray_extractor_ = GrayGeometryExtractor(n_jobs = self.n_jobs_, reduced = self.reduced_)


    def process_rgb_(self, rgb_images: numpy.ndarray, do_fit: bool, dump_name: typing.Optional[str] = None) -> numpy.ndarray:
        return utils.process_iter_dump(self.rgb_extractor_, rgb_images, do_fit, self.features_dump_(dump_name))
    
    def process_gray_(self, gray_images: numpy.ndarray, do_fit: bool, dump_name: typing.Optional[str] = None) -> numpy.ndarray:
        return utils.process_iter_dump(self.gray_extractor_, gray_images, do_fit, self.features_dump_(dump_name))



import numpy
import joblib

import skimage.color

import cvtda.logging


def rgb2hsv(images: numpy.ndarray, n_jobs: int = -1) -> numpy.ndarray:
    return numpy.stack(
        joblib.Parallel(n_jobs = n_jobs)(
            joblib.delayed(skimage.color.rgb2hsv)(img)
            for img in cvtda.logging.logger().pbar(images, desc = "rgb2hsv")
        )
    )



from .classify import classify
from .estimate_quality import estimate_quality
from .information_value import calculate_binary_information_value
from .information_value import calculate_information_value
from .information_value import InformationValueFeatureSelector
from .NNClassifier import NNClassifier



import abc
import typing

import numpy
import joblib
import itertools
import matplotlib.pyplot as plt

import cvtda.logging
import cvtda.neural_network


class BaseLearner:
    def __init__(self, n_jobs: int = -1):
        self.n_jobs_ = n_jobs

    
    @abc.abstractmethod
    def fit(self, train: cvtda.neural_network.Dataset, val: typing.Optional[cvtda.neural_network.Dataset]):
        pass

            
    def estimate_quality(
        self,
        dataset: cvtda.neural_network.Dataset,
        ax: typing.Optional[plt.Axes] = None
    ):
        def calculate_distance_(i: int, j: int):
            return (i, j, self.calculate_distance_(i, j, dataset))

        idxs = list(itertools.product(range(len(dataset)), range(len(dataset))))
        distances_flat = joblib.Parallel(n_jobs = self.n_jobs_)(
            joblib.delayed(calculate_distance_)(i, j)
            for i, j in cvtda.logging.logger().pbar(idxs, desc = "Calculating pairwise distances")
        )

        correct_dists, incorrect_dists = {}, {}
        for i, j, distance in cvtda.logging.logger().pbar(distances_flat, desc = "Analyzing distances"):
            label1, label2 = dataset.get_labels([ i, j ])
            if label1 == label2:
                correct_dists[(i, j)] = distance
            else:
                incorrect_dists[(i, j)] = distance
        correct_dists_values = list(correct_dists.values())
        incorrect_dists_values = list(incorrect_dists.values())
        
        if ax is not None:
            ax.set_ylim(0, 1)
            ax.get_yaxis().set_ticks([])
            ax.plot(
                correct_dists_values,
                numpy.ones_like(correct_dists_values) * 0.35,
                'x',
                label = "Одного человека"
            )
            ax.plot(
                incorrect_dists_values,
                numpy.ones_like(incorrect_dists_values) * 0.65,
                'x',
                label = "Разных людей"
            )
        return correct_dists, incorrect_dists


    @abc.abstractmethod
    def calculate_distance_(self, first: int, second: int, dataset: cvtda.neural_network.Dataset):
        pass


import os
import typing

import numpy

import cvtda.logging

from .base import BaseDumper

class NumpyDumper(BaseDumper[numpy.ndarray]):
    def __init__(self, directory: str):
        self.directory_ = directory

    def get_file_name_(self, name: str):
        return os.path.join(self.directory_, f"{name}.npy")
    
    def execute(self, function: typing.Callable[[typing.Any], numpy.ndarray], name: str, *function_args) -> numpy.ndarray:
        if self.has_dump(name):
            return self.get_dump(name)
        result = function(*function_args)
        self.save_dump(result, name)
        return result

    def save_dump(self, data: numpy.ndarray, name: str):
        file = self.get_file_name_(name)
        cvtda.logging.logger().print(f"Saving the result to {file}")
        os.makedirs(os.path.dirname(file), exist_ok = True)
        numpy.save(file, data)

    def has_dump(self, name: str) -> bool:
        return os.path.exists(self.get_file_name_(name))
    
    def get_dump_impl_(self, name: str) -> numpy.ndarray:
        file = self.get_file_name_(name)
        cvtda.logging.logger().print(f"Got the result from {file}")
        return numpy.load(self.get_file_name_(name))



import typing

import torch
import torch.utils.data
import pytorch_metric_learning.miners
import pytorch_metric_learning.losses
import pytorch_metric_learning.samplers
import pytorch_metric_learning.utils.accuracy_calculator

import cvtda.utils
import cvtda.logging
import cvtda.neural_network

from .BaseLearner import BaseLearner


class NNLearner(BaseLearner):
    def __init__(
        self,

        n_jobs: int = -1,
        random_state: int = 42,

        device: torch.device = torch.device("cuda"),
        batch_size: int = 64,
        learning_rate: float = 1e-4,
        n_epochs: int = 25,

        margin: int = 1,
        latent_dim: int = 256,
        
        skip_diagrams: bool = False,
        skip_images: bool = False,
        skip_features: bool = False
    ):
        super().__init__(n_jobs)
        self.random_state_ = random_state

        self.device_ = device
        self.batch_size_ = batch_size
        self.learning_rate_ = learning_rate
        self.n_epochs_ = n_epochs

        self.margin_ = margin
        self.latent_dim_ = latent_dim

        self.skip_diagrams_ = skip_diagrams
        self.skip_images_ = skip_images
        self.skip_features_ = skip_features


    def fit(self, train: cvtda.neural_network.Dataset, val: typing.Optional[cvtda.neural_network.Dataset]):
        cvtda.utils.set_random_seed(self.random_state_)

        train_mpc_sampler = pytorch_metric_learning.samplers.MPerClassSampler(
            m = 4,
            labels = train.labels,
            length_before_new_iter = self.batch_size_ * 20
        )
        train_dl = torch.utils.data.DataLoader(
            torch.utils.data.TensorDataset(train.images, train.features, train.labels, torch.arange(len(train))),
            batch_size = self.batch_size_,
            sampler = train_mpc_sampler
        )
        self.init_(*next(iter(train_dl)), train)

        train_miner = pytorch_metric_learning.miners.TripletMarginMiner(
            margin = self.margin_, 
            type_of_triplets = "all"
        )
        train_loss = pytorch_metric_learning.losses.TripletMarginLoss(margin = self.margin_)
        metrics = pytorch_metric_learning.utils.accuracy_calculator.AccuracyCalculator()
        
        if val is not None:
            val_dl = torch.utils.data.DataLoader(
            torch.utils.data.TensorDataset(val.images, val.features, val.labels, torch.arange(len(val))),
                batch_size = self.batch_size_
            )

        pbar = cvtda.logging.logger().pbar(range(self.n_epochs_), desc = "Train")
        for _ in pbar:
            sum_loss = 0

            self.model_list_.train()
            for images, features, labels, diagram_idxs in train_dl:
                self.optimizer_.zero_grad()
                embeddings = self.forward_(images, features, diagram_idxs, train)
                indices = train_miner(embeddings, labels)
                loss = train_loss(embeddings, labels, indices)
                loss.backward()
                self.optimizer_.step()
                sum_loss += loss.item()
            postfix = { 'loss': sum_loss }
            self.scheduler_.step()

            if val is not None:
                self.model_list_.eval()
                all_embeddings, all_targets = [], []
                for images, features, labels, diagram_idxs in val_dl:
                    with torch.no_grad():
                        all_targets.append(labels)
                        all_embeddings.append(self.forward_(images, features, diagram_idxs, val))
                result = metrics.get_accuracy(torch.cat(all_embeddings, dim = 0), torch.cat(all_targets, dim = 0))
                postfix = { **postfix, **result }

            cvtda.logging.logger().set_pbar_postfix(pbar, postfix)

        self.model_list_.eval()
        self.fitted_ = True
        return self
    
    def calculate_distance_(self, first: int, second: int, dataset: cvtda.neural_network.Dataset):
        self.model_list_.eval()
        idx = [ first, second ]
        embeddings = self.forward_(dataset.images[idx], dataset.features[idx], idx, dataset)
        return torch.sqrt(torch.sum((embeddings[0] - embeddings[1]) ** 2)).item()


    def init_(
        self, 
        images: torch.Tensor,
        features: torch.Tensor,
        labels: torch.Tensor,
        diagram_idxs: torch.Tensor,
        dataset: cvtda.neural_network.Dataset
    ):
        diagrams = [] if self.skip_diagrams_ else dataset.get_diagrams(diagram_idxs)

        is_no_topology = (self.skip_diagrams_ and self.skip_features_ and not self.skip_images_)
        print('Topology: ', not is_no_topology)
        print('Images: ', images.shape)
        print('Features: ', features.shape)

        if is_no_topology:
            images_output = len(torch.unique(labels))
        else:
            images_output = self.latent_dim_

        self.nn_base_ = cvtda.neural_network.NNBase(
            num_diagrams = len(diagrams) // 2,
            skip_diagrams = self.skip_diagrams_,
            skip_images = self.skip_images_,
            skip_features = self.skip_features_,
            images_n_channels = images.shape[1],
            images_output = images_output
        ).to(self.device_).train()

        self.model_ = torch.nn.Sequential(
            torch.nn.Dropout(0.3), torch.nn.LazyLinear(1024), torch.nn.BatchNorm1d(1024), torch.nn.GELU(),
            torch.nn.Dropout(0.2), torch.nn.Linear(1024, 768), torch.nn.BatchNorm1d(768), torch.nn.GELU(),
            torch.nn.Dropout(0.1), torch.nn.Linear(768, 512), torch.nn.BatchNorm1d(512), torch.nn.GELU(),
            torch.nn.Linear(512, self.latent_dim_)
        ).to(self.device_).train()

        if is_no_topology:
            self.model_ = torch.nn.Identity()

        self.model_list_ = torch.nn.ModuleList([ self.nn_base_, self.model_ ])

        self.optimizer_ = torch.optim.AdamW(
            params = self.model_list_.parameters(),
            lr = self.learning_rate_ * 100
        )
        
        def lr_scheduler_lambda(epoch):
            if epoch < self.n_epochs_ // 10:
                return 1
            if epoch < self.n_epochs_ // 4:
                return 0.1
            if epoch < self.n_epochs_ // 2:
                return 0.01
            if epoch < 3 * self.n_epochs_ // 4:
                return 0.001
            return 0.0001
        self.scheduler_ = torch.optim.lr_scheduler.LambdaLR(self.optimizer_, lr_scheduler_lambda)

        self.forward_(images, features, diagram_idxs, dataset)
        if not is_no_topology:
            cvtda.logging.logger().print(f'Input to LazyLinear: {self.model_[1].in_features}')
        cvtda.logging.logger().print(f'Parameters: {sum(p.numel() for p in self.model_list_.parameters())}')

    def forward_(
        self,
        images: torch.Tensor,
        features: torch.Tensor,
        diagram_idxs: torch.Tensor,
        dataset: cvtda.neural_network.Dataset
    ) -> torch.Tensor:
        images = images.to(self.device_)
        features = features.to(self.device_)
        if self.skip_diagrams_:
            return self.model_(self.nn_base_(images, features))
        return self.model_(self.nn_base_(images, features, *dataset.get_diagrams(diagram_idxs)))



import os
import random

import numpy
import torch

def set_random_seed(seed: int = 42):
    random.seed(seed)
    numpy.random.seed(seed)
    torch.manual_seed(seed)
    torch.cuda.manual_seed(seed)
    os.environ["PYTHONHASHSEED"] = str(seed)
    torch.backends.cudnn.deterministic = True



import typing

import numpy
import itertools
import gtda.images
import sklearn.base
import gtda.homology

import cvtda.utils
import cvtda.logging

from . import utils
import cvtda.dumping
from .interface import TopologicalExtractor


class FiltrationExtractor(TopologicalExtractor):
    def __init__(
        self,
        filtration_class,
        filtation_kwargs: dict,
        binarizer_threshold: float,

        n_jobs: int = -1,
        reduced: bool = True,
        only_get_from_dump: bool = False,
        return_diagrams: bool = False,
        **kwargs
    ):
        super().__init__(
            supports_rgb = False,
            n_jobs = n_jobs,
            reduced = reduced,
            only_get_from_dump = only_get_from_dump,
            return_diagrams = return_diagrams,
            filtration_class = filtration_class,
            filtation_kwargs = filtation_kwargs,
            binarizer_threshold = binarizer_threshold,
            **kwargs
        )

        self.binarizer_ = gtda.images.Binarizer(threshold = binarizer_threshold, n_jobs = self.n_jobs_)
        self.filtration_ = filtration_class(**filtation_kwargs, n_jobs = self.n_jobs_)
        self.persistence_ = gtda.homology.CubicalPersistence(homology_dimensions = [ 0, 1 ], n_jobs = self.n_jobs_)


    def get_diagrams_(self, images: numpy.ndarray, do_fit: bool, dump_name: typing.Optional[str] = None):
        cvtda.logging.logger().print(f"FiltrationExtractor: processing {dump_name}, do_fit = {do_fit}, filtration = {self.filtration_}")
        
        bin_images = utils.process_iter(self.binarizer_, images, do_fit)
        assert bin_images.shape == images.shape

        filtrations = utils.process_iter(self.filtration_, bin_images, do_fit)
        assert filtrations.shape == images.shape

        return utils.process_iter_dump(self.persistence_, filtrations, do_fit, self.diagrams_dump_(dump_name))


class FiltrationsExtractor(sklearn.base.TransformerMixin):
    def __init__(
        self,

        n_jobs: int = -1,
        reduced: bool = True,
        only_get_from_dump: bool = False,
        return_diagrams: bool = False,

        binarizer_thresholds: typing.Optional[typing.List[float]] = None,
        height_filtration_directions: typing.Iterable[typing.Tuple[float, float]] = [
            [ -1, -1 ], [ 1, 1 ], [ 1, -1 ], [ -1, 1 ],
            [ 0, -1 ], [ 0, 1 ], [ -1, 0 ], [ 1, 0 ]
        ],
        num_radial_filtrations: int = 4,
        density_filtration_radiuses: typing.Iterable[int] = [ 1, 3 ],
    ):
        if not binarizer_thresholds:
            if reduced:
                binarizer_thresholds = [ 0.2, 0.4, 0.6 ]
            else:
                binarizer_thresholds = [ 0.1, 0.2, 0.3, 0.4, 0.5, 0.6, 0.7, 0.8, 0.9 ]

        self.fitted_ = False
        self.reduced_ = reduced
        self.return_diagrams_ = return_diagrams
        self.filtrations_kwargs_ = {
            'n_jobs': n_jobs,
            'reduced': reduced,
            'only_get_from_dump': only_get_from_dump,
            'return_diagrams': return_diagrams
        }

        self.binarizer_thresholds_ = binarizer_thresholds
        self.height_filtration_directions_ = height_filtration_directions
        self.num_radial_filtrations_ = num_radial_filtrations
        self.density_filtration_radiuses_ = density_filtration_radiuses

        self.filtration_extractors_: typing.List[typing.Tuple[FiltrationExtractor, str]] = []


    def fit(self, images: numpy.ndarray, dump_name: typing.Optional[str] = None):
        assert len(images.shape) >= 3, f'{len(images.shape) - 1}d images are not supported'
        cvtda.logging.logger().print("Fitting filtrations")

        self._fill_filtrations(images.shape[1], images.shape[2])
        for i, (filtration_extractor, name) in enumerate(self.filtration_extractors_):
            cvtda.logging.logger().print(f"Fitting filtration {i + 1}/{len(self.filtration_extractors_)}: {name}")
            filtration_extractor.fit(images, cvtda.dumping.dump_name_concat(dump_name, name))
        self.fitted_ = True
        return self
    
    def transform(self, images: numpy.ndarray, dump_name: typing.Optional[str] = None) -> numpy.ndarray:
        assert self.fitted_ is True, 'fit() must be called before transform()'
        cvtda.logging.logger().print("Applying filtrations")
        
        outputs = [ ]
        for i, (filtration_extractor, name) in enumerate(self.filtration_extractors_):
            cvtda.logging.logger().print(f"Applying filtration {i + 1}/{len(self.filtration_extractors_)}: {name}")
            outputs.append(filtration_extractor.transform(images, cvtda.dumping.dump_name_concat(dump_name, name)))
        return utils.hstack(outputs, not self.return_diagrams_)
    
    def fit_transform(self, images: numpy.ndarray, dump_name: typing.Optional[str] = None) -> numpy.ndarray:
        return self.fit(images, dump_name = dump_name).transform(images, dump_name = dump_name)


    def _fill_filtrations(self, width: int, height: int):
        radial_x = cvtda.utils.spread_points(width, self.num_radial_filtrations_)
        radial_y = cvtda.utils.spread_points(height, self.num_radial_filtrations_)
        cvtda.logging.logger().print(f"Calculated radial centers for images of size {width}x{height}: {radial_x}x{radial_y}")

        self.filtration_extractors_ = [ ]
        for binarizer_threshold in self.binarizer_thresholds_:
            self._add_height_filtrations(binarizer_threshold)
            self._add_radial_filtrations(binarizer_threshold, radial_x, radial_y)
            self._add_dilation_filtrations(binarizer_threshold)
            self._add_erosion_filtrations(binarizer_threshold)
            self._add_signed_distance_filtrations(binarizer_threshold)
            self._add_density_filtrations(binarizer_threshold)

    def _add_height_filtrations(self, binarizer_threshold: float):
        for direction in self.height_filtration_directions_:
            name = f'{int(binarizer_threshold * 10)}/HeightFiltrartion_{direction[0]}_{direction[1]}'
            extractor = FiltrationExtractor(
                gtda.images.HeightFiltration, { 'direction': numpy.array(direction) }, binarizer_threshold, **self.filtrations_kwargs_
            )
            self.filtration_extractors_.append((extractor, name))
            
    def _add_radial_filtrations(self, binarizer_threshold: float, radial_x: typing.List[int], radial_y: typing.List[int]):
        for center in list(itertools.product(radial_x, radial_y)):
            name = f'{int(binarizer_threshold * 10)}/RadialFiltration_{center[0]}_{center[1]}'
            extractor = FiltrationExtractor(
                gtda.images.RadialFiltration, { 'center': numpy.array(center) }, binarizer_threshold, **self.filtrations_kwargs_
            )
            self.filtration_extractors_.append((extractor, name))

    def _add_dilation_filtrations(self, binarizer_threshold: float):
        if self.reduced_:
            return
        name = f'{int(binarizer_threshold * 10)}/DilationFiltration'
        extractor = FiltrationExtractor(gtda.images.DilationFiltration, { }, binarizer_threshold, **self.filtrations_kwargs_)
        self.filtration_extractors_.append((extractor, name))

    def _add_erosion_filtrations(self, binarizer_threshold: float):
        if self.reduced_:
            return
        name = f'{int(binarizer_threshold * 10)}/ErosionFiltration'
        extractor = FiltrationExtractor(gtda.images.ErosionFiltration, { }, binarizer_threshold, **self.filtrations_kwargs_)
        self.filtration_extractors_.append((extractor, name))

    def _add_signed_distance_filtrations(self, binarizer_threshold: float):
        if self.reduced_:
            return
        name = f'{int(binarizer_threshold * 10)}/SignedDistanceFiltration'
        extractor = FiltrationExtractor(gtda.images.SignedDistanceFiltration, { }, binarizer_threshold, **self.filtrations_kwargs_)
        self.filtration_extractors_.append((extractor, name))
        
    def _add_density_filtrations(self, binarizer_threshold: float):
        if self.reduced_:
            return
        for radius in self.density_filtration_radiuses_:
            name = f'{int(binarizer_threshold * 10)}/DensityFiltration_{radius}'
            extractor = FiltrationExtractor(gtda.images.DensityFiltration, { 'radius': radius }, binarizer_threshold, **self.filtrations_kwargs_)
            self.filtration_extractors_.append((extractor, name))



import typing

import numpy
import joblib
import sklearn.base

import cvtda.logging

class DuplicateFeaturesRemover(sklearn.base.TransformerMixin):
    def __init__(
        self,
        n_jobs: int = -1,

        tolerance: float = 1e-8,

        partition_search_batch_size: int = 2000,
        naive_threshold: int = 300
    ):
        self.fitted_ = False
        self.n_jobs_ = n_jobs

        self.tolerance_ = tolerance

        self.partition_search_batch_size_ = partition_search_batch_size
        self.naive_threshold_ = naive_threshold
    
    def fit(self, features: numpy.ndarray):
        self.non_duplicates_ = self.analyze_columns_(features)[0]
        self.fitted_ = True
        return self
    
    def transform(self, features: numpy.ndarray) -> numpy.ndarray:
        assert self.fitted_ is True, 'fit() must be called before transform()'
        return features[:, self.non_duplicates_]


    def naive_find_duplicate_columns_(self, features: numpy.ndarray) -> set:
        duplicates = set()
        features = features.transpose()
        for i in range(features.shape[0]):
            if i in duplicates:
                continue
            comp = numpy.abs(features - features[i]) < self.tolerance_
            comp = numpy.sum(comp, axis = 1)
            comp = numpy.where(comp == features.shape[1])[0]
            for j in comp[1:]:
                duplicates.add(j)
        return duplicates

    def analyze_columns_(
        self,
        features: numpy.ndarray,
        force_naive: bool = False,
        depth: int = 0
    ) -> typing.List[int]:
        if (features.shape[1] < self.naive_threshold_) or force_naive:
            if force_naive:
                cvtda.logging.logger().print(f"Naive check for {features.shape[1]} features at depth {depth}")
            duplicates = self.naive_find_duplicate_columns_(features)
            return list(set(range(features.shape[1])) - duplicates), list(duplicates)
        
        partition_by = self.find_best_partition_(features)
        partition_item = features[partition_by, :]
        duplicates = set()

        prev_value = numpy.min(partition_item) - 5 * self.tolerance_
        pbar = cvtda.logging.logger().pbar(numpy.unique(partition_item))
        for partition_value in pbar:
            if partition_value - prev_value < self.tolerance_:
                continue

            partition_idxs = numpy.where(numpy.abs(partition_item - partition_value) < self.tolerance_)[0]
            partition_idxs = numpy.setdiff1d(partition_idxs, numpy.array(list(duplicates)), assume_unique = True)

            cvtda.logging.logger().set_pbar_postfix(pbar, {
                'partition_by': partition_by,
                'num_features': len(partition_idxs),
                'duplicates': len(duplicates)
            })

            is_bad_partition = (len(partition_idxs) == features.shape[1])
            with cvtda.logging.DevNullLogger():
                _, partition_duplicates = self.analyze_columns_(
                    features[:, partition_idxs],
                    force_naive = is_bad_partition,
                    depth = depth + 1
                )

            for item in partition_idxs[list(partition_duplicates)]:
                duplicates.add(item)
            prev_value = partition_value

        cvtda.logging.logger().print(f'Found {len(duplicates)} duplicates')
        return list(set(range(features.shape[1])) - duplicates), list(duplicates)

    def find_best_partition_(self, features: numpy.ndarray) -> int:
        def bulk_biggest_partition_(items: numpy.ndarray) -> list:
            return [ numpy.max(numpy.unique(item, return_counts = True)[1]) for item in items ]
        
        partitions = joblib.Parallel(return_as = 'generator', n_jobs = self.n_jobs_)(
            joblib.delayed(bulk_biggest_partition_)(features[batch_start:batch_start + self.partition_search_batch_size_])
            for batch_start in range(0, features.shape[0], self.partition_search_batch_size_)
        )
        partitions = numpy.hstack(list(partitions))
        return numpy.argmin(partitions)



import numpy

def spread_points(size: int, num_points: int):
    assert size % 2 == 0, 'Odd size is not implemented'
    
    center, spacing = size // 2, size // (num_points + 1)
    centers = numpy.array(range((spacing + 1) // 2, center, spacing + 1))
    centers = numpy.array([ *(center - centers - 1), *(center + centers) ])
    if num_points % 2 == 1:
        centers = numpy.array([ *centers, center ])
    centers.sort()
    return centers



import typing

import torch

import cvtda.neural_network
from .BaseLearner import BaseLearner


class SimpleTopologicalLearner(BaseLearner):
    def fit(self, train: cvtda.neural_network.Dataset, val: typing.Optional[cvtda.neural_network.Dataset]):
        pass

    def calculate_distance_(self, first: int, second: int, dataset: cvtda.neural_network.Dataset):
        return torch.sqrt(torch.sum((dataset.features[first] - dataset.features[second]) ** 2))



import typing

import numpy
import gtda.images
import sklearn.base
import sklearn.preprocessing

from . import utils
import cvtda.logging
import cvtda.dumping
from .FiltrationsExtractor import FiltrationsExtractor
from .GreyscaleExtractor import GreyscaleExtractor
from .PointCloudsExtractor import PointCloudsExtractor
from .GeometryExtractor import GeometryExtractor


class FeatureExtractor(sklearn.base.TransformerMixin):
    def __init__(
        self,
        n_jobs: int = -1,
        reduced: bool = True,
        only_get_from_dump: bool = False,
        return_diagrams: bool = False,

        num_radial_filtrations: int = 4,
        binarizer_thresholds: typing.Optional[typing.List[float]] = None,
    ):
        self.fitted_ = False
        self.reduced_ = reduced
        self.return_diagrams_ = return_diagrams
        
        extractor_kwargs = { 'n_jobs': n_jobs, 'reduced': reduced, 'only_get_from_dump': only_get_from_dump }
        topological_extractor_kwargs = { **extractor_kwargs, 'return_diagrams': return_diagrams }

        self.inverter_ = gtda.images.Inverter()

        self.point_clouds_extractor_ = PointCloudsExtractor(**topological_extractor_kwargs)
        self.greyscale_extractor_ = GreyscaleExtractor(**topological_extractor_kwargs)
        self.inverted_greyscale_extractor_ = GreyscaleExtractor(**topological_extractor_kwargs)
        self.filtrations_extractor_ = FiltrationsExtractor(
            **topological_extractor_kwargs,
            binarizer_thresholds = binarizer_thresholds,
            num_radial_filtrations = num_radial_filtrations
        )
        self.geometry_extractor_ = GeometryExtractor(**extractor_kwargs)

        self.scaler_ = sklearn.preprocessing.StandardScaler(copy = False)

    def fit(self, images: numpy.ndarray, dump_name: typing.Optional[str] = None):
        self.process_(images, do_fit = True, dump_name = dump_name)
        self.fitted_ = True
        return self
    
    def transform(self, images: numpy.ndarray, dump_name: typing.Optional[str] = None):
        assert self.fitted_ is True, 'fit() must be called before transform()'
        return self.process_(images, do_fit = False, dump_name = dump_name)
    
    def fit_transform(self, images: numpy.ndarray, dump_name: typing.Optional[str] = None):
        return self.fit(images, dump_name = dump_name).transform(images, dump_name = dump_name)


    def process_(self, images: numpy.ndarray, do_fit: bool, dump_name: typing.Optional[str] = None):
        cvtda.logging.logger().print("Calculating inverted images")
        inverted_images = utils.process_iter(self.inverter_, images, do_fit = do_fit)

        point_clouds_dump = cvtda.dumping.dump_name_concat(dump_name, "point_clouds")
        greyscale_dump = cvtda.dumping.dump_name_concat(dump_name, "greyscale")
        inverted_greyscale_dump = cvtda.dumping.dump_name_concat(dump_name, "inverted_greyscale")
        filtrations_dump = cvtda.dumping.dump_name_concat(dump_name, "filtrations")
        geometry_dump = cvtda.dumping.dump_name_concat(dump_name, "geometry")

        results = []
        if not self.reduced_:
            results.append(utils.process_iter(self.point_clouds_extractor_, images, do_fit, point_clouds_dump))

        results.append(utils.process_iter(self.greyscale_extractor_, images, do_fit, greyscale_dump))
        results.append(utils.process_iter(self.inverted_greyscale_extractor_, inverted_images, do_fit, inverted_greyscale_dump))
        results.append(utils.process_iter(self.filtrations_extractor_, images, do_fit, filtrations_dump))
        
        if not self.return_diagrams_:
            results.append(utils.process_iter(self.geometry_extractor_, images, do_fit, geometry_dump))
        
        results = utils.hstack(results, not self.return_diagrams_)
        if self.return_diagrams_:
            cvtda.logging.logger().print("Diagrams requested. Returning diagrams.")
            return results
        
        cvtda.logging.logger().print("Applying StandardScaler.")
        return utils.process_iter(self.scaler_, results, do_fit)



from .base import BaseDumper
from .numpy import NumpyDumper
from .devnull import DevNullDumper
from .dump_name_concat import dump_name_concat

def dumper() -> BaseDumper:
    if BaseDumper.current_dumper is None:
        BaseDumper.current_dumper = NumpyDumper("./")
    return BaseDumper.current_dumper



import os
import typing

import numpy
import torch
import pandas
import xgboost
import catboost
import sklearn.base
import sklearn.ensemble
import sklearn.neighbors
import matplotlib.pyplot as plt

import cvtda.dumping
import cvtda.neural_network
from .NNClassifier import NNClassifier
from .estimate_quality import estimate_quality

def classify(
    train_images: numpy.ndarray,
    train_features: numpy.ndarray,
    train_labels: numpy.ndarray,
    train_diagrams: typing.Optional[typing.List[numpy.ndarray]],

    test_images: numpy.ndarray,
    test_features: numpy.ndarray,
    test_labels: numpy.ndarray,
    test_diagrams: typing.Optional[typing.List[numpy.ndarray]],

    label_names: typing.Optional[typing.List[str]] = None,
    confusion_matrix_include_values: bool = True,

    n_jobs: int = -1,
    random_state: int = 42,
    dump_name: typing.Optional[str] = None,
    only_get_from_dump: bool = False,

    knn_neighbours: int = 50,

    random_forest_estimators: int = 100,

    nn_device: torch.device = torch.device('cuda'),
    nn_batch_size: int = 256,
    nn_learning_rate: float = 1e-3,
    nn_epochs: int = 100,

    grad_boost_max_iter: int = 20,
    grad_boost_max_depth: int = 4,
    grad_boost_max_features: float = 0.1,

    xgboost_n_classifiers: int = 25,
    xgboost_max_depth: int = 4,
    xgboost_device: str = 'gpu',

    catboost_iterations: int = 600,
    catboost_depth: int = 4,
    catboost_device: str = 'GPU',

    without_diagrams: bool = False
):
    if (train_images is not None) and (not only_get_from_dump):
        nn_train = cvtda.neural_network.Dataset(
            train_images, train_diagrams, train_features, train_labels, n_jobs = n_jobs, device = nn_device
        )
        nn_test = cvtda.neural_network.Dataset(
            test_images, test_diagrams, test_features, test_labels, n_jobs = n_jobs, device = nn_device
        )

    def classify_one(classifier: sklearn.base.ClassifierMixin, name: str, display_name: str, ax: plt.Axes):
        if without_diagrams and name == 'NNClassifier_diagrams':
            print(f'Skipping {name} - {classifier}')
            return {}

        print(f'Trying {name} - {classifier}')

        dumper = cvtda.dumping.dumper()
        model_dump_name = cvtda.dumping.dump_name_concat(dump_name, name)
        if only_get_from_dump or dumper.has_dump(model_dump_name):
            y_pred_proba = dumper.get_dump(model_dump_name)
        else:
            if type(classifier) == NNClassifier:
                classifier.fit(nn_train, nn_test)
                y_pred_proba = classifier.predict_proba(nn_test)
            else:
                classifier.fit(train_features, train_labels)
                y_pred_proba = classifier.predict_proba(test_features)

            if model_dump_name is not None:
                dumper.save_dump(y_pred_proba, model_dump_name)
                
        ax.set_title(display_name)
        result = {
            'classifier': display_name,
            **estimate_quality(
                y_pred_proba, test_labels, ax,
                label_names = label_names, confusion_matrix_include_values = confusion_matrix_include_values
            )
        }
        print(result)
        return result

    classifiers = [
        sklearn.neighbors.KNeighborsClassifier(
            n_jobs = n_jobs,
            n_neighbors = knn_neighbours
        ),
        sklearn.ensemble.RandomForestClassifier(
            n_estimators = random_forest_estimators,
            random_state = random_state,
            n_jobs = n_jobs
        ),
        sklearn.ensemble.HistGradientBoostingClassifier(
            random_state = random_state,
            max_iter = grad_boost_max_iter,
            max_depth = grad_boost_max_depth,
            max_features = grad_boost_max_features,
            verbose = 2
        ),
        catboost.CatBoostClassifier(
            iterations = catboost_iterations,
            depth = catboost_depth,
            random_seed = random_state,
            loss_function = 'MultiClass',
            devices = '0-3',
            task_type = catboost_device,
            verbose = True
        ),
        xgboost.XGBClassifier(
            n_jobs = n_jobs,
            n_estimators = xgboost_n_classifiers,
            max_depth = xgboost_max_depth,
            device = xgboost_device
        ),
        NNClassifier(
            random_state = random_state,
            device = nn_device,
            batch_size = nn_batch_size,
            learning_rate = nn_learning_rate,
            n_epochs = nn_epochs * 2,
            skip_diagrams = True,
            skip_images = True,
            skip_features = False,
        ),
        NNClassifier(
            random_state = random_state,
            device = nn_device,
            batch_size = nn_batch_size,
            learning_rate = nn_learning_rate,
            n_epochs = nn_epochs // 5,
            skip_diagrams = False,
            skip_images = True,
            skip_features = True,
        ),
        NNClassifier(
            random_state = random_state,
            device = nn_device,
            batch_size = nn_batch_size,
            learning_rate = nn_learning_rate,
            n_epochs = nn_epochs,
            skip_diagrams = True,
            skip_images = False,
            skip_features = True,
        ),
        NNClassifier(
            random_state = random_state,
            device = nn_device,
            batch_size = nn_batch_size,
            learning_rate = nn_learning_rate,
            n_epochs = nn_epochs,
            skip_diagrams = True,
            skip_images = False,
            skip_features = False,
        )
    ]
    names = [
        'KNeighborsClassifier',
        'RandomForestClassifier',
        'HistGradientBoostingClassifier',
        'CatBoostClassifier',
        'XGBClassifier',
        'NNClassifier_features',
        'NNClassifier_diagrams',
        'NNClassifier_images',
        'NNClassifier_features_images'
    ]
    display_names = [
        'Метод k ближайших соседей',
        'Случайный лес',
        'Град. бустинг на основе гистограмм',
        'CatBoost',
        'XGBoost',
        'Нейронная сеть для тополог. признаков',
        'Обучаемая векторизация диаграмм',
        'ResNet50 – базовая модель',
        'Комбинированная нейронная сеть'
    ]

    figure, axes = plt.subplots(3, 3, figsize = (15, 15))
    df = pandas.DataFrame([ classify_one(*args) for args in zip(classifiers, names, display_names, axes.flat) ])
    figure.tight_layout()

    dumper = cvtda.dumping.dumper()
    if (dump_name is not None) and isinstance(dumper, cvtda.dumping.NumpyDumper):
        file = dumper.get_file_name_(cvtda.dumping.dump_name_concat(dump_name, "confusion_matrixes"))
        os.makedirs(os.path.dirname(file), exist_ok = True)
        figure.savefig(file[:-4] + ".svg")
        figure.savefig(file[:-4] + ".png")
        df.to_csv(dumper.get_file_name_(cvtda.dumping.dump_name_concat(dump_name, "quality_metrics.csv"))[:-4])
    return df



import abc
import typing

import numpy
import gtda.diagrams

from .. import utils
import cvtda.logging
import cvtda.dumping
from .Extractor import Extractor
from ..DiagramVectorizer import DiagramVectorizer

class TopologicalExtractor(Extractor):
    def __init__(
        self,
        supports_rgb: bool,
        n_jobs: int = -1,
        reduced: bool = True,
        only_get_from_dump: bool = False,
        topo_only_get_from_dump: bool = False,
        return_diagrams: bool = False,
        **kwargs
    ):
        super().__init__(
            n_jobs = n_jobs,
            reduced = reduced,
            only_get_from_dump = False,
            return_diagrams = return_diagrams,
            topo_only_get_from_dump = (topo_only_get_from_dump or only_get_from_dump),
            **kwargs
        )

        self.topo_only_get_from_dump_ = (topo_only_get_from_dump or only_get_from_dump)
        self.return_diagrams_ = return_diagrams
        self.supports_rgb_ = supports_rgb

        self.vectorizer_ = DiagramVectorizer(n_jobs = self.n_jobs_, reduced = self.reduced_)
        self.scaler_ = gtda.diagrams.Scaler(n_jobs = self.n_jobs_)
        

    def final_dump_name_(self, dump_name: typing.Optional[str] = None):
        return None
    
    def diagrams_dump_(self, dump_name: typing.Optional[str]):
        return cvtda.dumping.dump_name_concat(dump_name, "diagrams")
    
    def force_numpy_(self):
        return not self.return_diagrams_


    def process_rgb_(self, rgb_images: numpy.ndarray, do_fit: bool, dump_name: typing.Optional[str] = None):
        if not self.supports_rgb_:
            if self.return_diagrams_:
                return []
            else:
                return numpy.empty((len(rgb_images), 0))
        return self.do_work_(rgb_images, do_fit, dump_name)

    def process_gray_(self, gray_images: numpy.ndarray, do_fit: bool, dump_name: typing.Optional[str] = None):
        return self.do_work_(gray_images, do_fit, dump_name)
    
    def do_work_(self, images: numpy.ndarray, do_fit: bool, dump_name: typing.Optional[str] = None):
        if self.topo_only_get_from_dump_:
            if self.return_diagrams_:
                diagrams = cvtda.dumping.dumper().get_dump(self.diagrams_dump_(dump_name))
                cvtda.logging.logger().print("Applying Scaler to persistence diagrams.")
                return utils.process_iter(self.scaler_, diagrams, do_fit)
            else:
                return cvtda.dumping.dumper().get_dump(self.features_dump_(dump_name))

        diagrams = self.get_diagrams_(images, do_fit, dump_name)
        cvtda.logging.logger().print("Applying Scaler to persistence diagrams.")
        diagrams = utils.process_iter(self.scaler_, diagrams, do_fit)
        if self.return_diagrams_:
            return diagrams
        return utils.process_iter_dump(self.vectorizer_, diagrams, do_fit, self.features_dump_(dump_name))
    
    @abc.abstractmethod
    def get_diagrams_(self, images: numpy.ndarray, do_fit: bool, dump_name: typing.Optional[str] = None):
        pass


import typing

from .base import BaseDumper

T = typing.TypeVar("T")

class DevNullDumper(BaseDumper[T]):
    def execute(self, function: typing.Callable[[typing.Any], T], name: str, *function_args) -> T:
        return function(*function_args)

    def save_dump(self, data: T, name: str):
        return

    def has_dump(self, name: str) -> bool:
        return False
    
    def get_dump_impl_(self, name: str) -> T:
        raise NotImplementedError



import typing

from .base import BaseLogger

T = typing.TypeVar("T")

class DevNullLogger(BaseLogger):
    def print(self, data: T) -> None:
        pass

    def pbar(
        self,
        data: typing.Iterable[T],
        total: int = None,
        desc: typing.Optional[str] = None
    ) -> typing.Iterable[T]:
        return data

    def zip(
        self,
        *iterables, 
        desc: typing.Optional[str] = None
    ):
        return zip(*iterables)

    def set_pbar_postfix(self, pbar: typing.Any, data: dict):
        pass



import numpy
import pandas
import joblib
import sklearn.base
import matplotlib.pyplot as plt

import cvtda.logging


def calculate_binary_information_value(feature: numpy.ndarray, target: numpy.ndarray, bins: int) -> float:
    quantiles = numpy.linspace(0, 1, bins + 1)
    bins = numpy.unique(numpy.quantile(feature, quantiles))
    bins[0] -= 0.0001
    bins[-1] += 0.0001
    
    feature = numpy.digitize(feature, bins, right = True)
    n, events = [ ], [ ]
    for bin in range(1, len(bins)):
        mask = (feature == bin)
        n.append(mask.sum())
        events.append((target * mask).sum())
    n = numpy.array(n)
    events = numpy.array(events)

    non_events = n - events
    events_prc = numpy.maximum(events, 0.5) / events.sum()
    non_events_prc = numpy.maximum(non_events, 0.5) / non_events.sum()

    woe = numpy.log(events_prc / non_events_prc)
    iv = woe * (events_prc - non_events_prc)
    return iv.sum()


def calculate_information_value_one_feature(
    feature: numpy.ndarray,
    y_true: numpy.ndarray,
    bins: int
) -> dict:
    IVs = []
    for class_idx in range(numpy.max(y_true)):
        target = (y_true == class_idx).astype(int)
        IVs.append(calculate_binary_information_value(feature, target, bins))
    return numpy.mean(IVs)


def calculate_information_value(
    features: numpy.ndarray,
    y_true: numpy.ndarray,
    bins: int = 10,
    n_jobs: int = -1
) -> pandas.DataFrame:
    iv_generator = joblib.Parallel(return_as = 'generator', n_jobs = n_jobs)(
        joblib.delayed(calculate_information_value_one_feature)(
            features[:, idx],
            y_true,
            bins
        ) for idx in range(features.shape[1])
    )
    pbar = cvtda.logging.logger().pbar(iv_generator, total = features.shape[1], desc = 'information values')
    return numpy.array(list(pbar))


class InformationValueFeatureSelector(sklearn.base.TransformerMixin):
    def __init__(
        self,
        n_jobs: int = -1,
        verbose: bool = True,

        bins: int = 10,
        threshold: float = 0.5
    ):
        self.fitted_ = False
        self.n_jobs_ = n_jobs

        self.bins_ = bins
        self.threshold_ = threshold

    def fit(self, features: numpy.ndarray, target: numpy.ndarray):
        cvtda.logging.logger().print('Fitting the information value feature selector')
        self.IV_ = calculate_information_value(
            features,
            target,
            bins = self.bins_,
            n_jobs = self.n_jobs_
        )
        self.good_features_idx_ = numpy.where(self.IV_ > self.threshold_)[0]
        
        cvtda.logging.logger().print('Fitting complete')
        self.fitted_ = True
        return self
    
    def transform(self, features: numpy.ndarray) -> numpy.ndarray:
        assert self.fitted_ is True, 'fit() must be called before transform()'
        return features[:, self.good_features_idx_]

    def hist(self, bins: int = 50):
        assert self.fitted_ is True, 'fit() must be called before hist()'
        return plt.hist(self.IV_, bins = bins)



import numpy
import numpy.ma.core


def base_(data: numpy.ndarray) -> numpy.ndarray:
    if type(data) == numpy.ma.core.MaskedArray:
        return numpy.ma.hstack([
            numpy.ma.sum(numpy.ma.abs(data), axis = 1, keepdims = True), # manhattan norm
            numpy.ma.sqrt(numpy.ma.sum(data ** 2, axis = 1, keepdims = True)), # euclidean norm
            numpy.ma.max(numpy.ma.abs(data), axis = 1, keepdims = True), # infinity norm
        ]).filled(0)
    else:
        return numpy.hstack([
            numpy.trapz(data, axis = 1).reshape(-1, 1), # integral
            numpy.sum(numpy.abs(data), axis = 1, keepdims = True), # manhattan norm
            numpy.sqrt(numpy.sum(data ** 2, axis = 1, keepdims = True)), # euclidean norm
            numpy.max(numpy.abs(data), axis = 1, keepdims = True), # infinity norm
        ])

def extension_(data: numpy.ndarray) -> numpy.ndarray:
    if type(data) == numpy.ma.core.MaskedArray:
        return numpy.ma.hstack([
            numpy.ma.max(data, axis = 1, keepdims = True),
            numpy.ma.sum(data, axis = 1, keepdims = True),
            numpy.ma.mean(data, axis = 1, keepdims = True),
            numpy.ma.std(data, axis = 1, keepdims = True),
            numpy.ma.median(data, axis = 1, keepdims = True),
        ]).filled(0)
    else:
        return numpy.hstack([
            numpy.max(data, axis = 1, keepdims = True),
            numpy.sum(data, axis = 1, keepdims = True),
            numpy.mean(data, axis = 1, keepdims = True),
            numpy.std(data, axis = 1, keepdims = True),
            numpy.median(data, axis = 1, keepdims = True),
        ])
    
    
def sequence2features(sequence_batch: numpy.ndarray, reduced: bool = True) -> numpy.ndarray:
    base = base_(sequence_batch)
    if reduced:
        return base
    
    if type(sequence_batch) == numpy.ma.core.MaskedArray:
        return numpy.ma.hstack([ base, extension_(sequence_batch) ])
    else:
        return numpy.hstack([ base, extension_(sequence_batch) ])



import typing

import numpy
import joblib
import sklearn.base
import gtda.diagrams

import cvtda.utils
import cvtda.logging


class DiagramVectorizer(sklearn.base.TransformerMixin):
    def __init__(
        self,
        n_jobs: int = -1,
        reduced: bool = True,

        n_bins: int = 64,
        batch_size: int = 512,
        filtering_percentile: int = 10,
        
        persistence_landscape_layers: int = 3,
        silhouette_powers: typing.List[int] = [ 1, 2 ],
        heat_kernel_sigmas: typing.List[float] = [ 0.1, 1.0, numpy.pi ],
        persistence_image_sigmas: typing.List[float] = [ 0.1, 1.0, numpy.pi ]
    ):
        self.fitted_ = False
        self.n_jobs_ = n_jobs
        self.reduced_ = reduced
        self.batch_size_ = batch_size
        self.filtering_percentile_ = filtering_percentile
        
        self.betti_curve_ = gtda.diagrams.BettiCurve(n_bins = n_bins, n_jobs = 1)

        self.persistence_landscape_ = gtda.diagrams.PersistenceLandscape(
            n_layers = persistence_landscape_layers,
            n_bins = n_bins,
            n_jobs = 1
        )

        self.silhouettes_ = [
            gtda.diagrams.Silhouette(power = power, n_bins = n_bins, n_jobs = 1)
            for power in silhouette_powers
        ]
        
        self.persistence_entropy_ = gtda.diagrams.PersistenceEntropy(nan_fill_value = 0, n_jobs = 1)
        
        self.number_of_points_ = gtda.diagrams.NumberOfPoints(n_jobs = 1)
        
        self.heat_kernels_ = [
            gtda.diagrams.HeatKernel(sigma = sigma, n_bins = n_bins, n_jobs = 1)
            for sigma in heat_kernel_sigmas
        ]
        
        self.persistence_images_ = [
            gtda.diagrams.PersistenceImage(sigma = sigma, n_bins = n_bins, n_jobs = 1)
            for sigma in persistence_image_sigmas
        ]


    def fit(self, diagrams: numpy.ndarray):
        self.homology_dimensions_ = numpy.unique(diagrams[:, :, 2])
        
        self.filtering_epsilon_ = self.determine_filtering_epsilon_(diagrams)
        self.filtering_ = gtda.diagrams.Filtering(epsilon = self.filtering_epsilon_).fit(diagrams)

        self.betti_curve_.fit(diagrams)
        
        self.persistence_landscape_.fit(diagrams)
        
        for silhouette in self.silhouettes_:
            silhouette.fit(diagrams)
        
        self.persistence_entropy_.fit(diagrams)
        
        self.number_of_points_.fit(diagrams)

        for heat_kernel in self.heat_kernels_:
            heat_kernel.fit(diagrams)
            
        for persistence_image in self.persistence_images_:
            persistence_image.fit(diagrams)

        cvtda.logging.logger().print('DiagramVectorizer: fitting complete')
        self.fitted_ = True
        return self
    
    def transform(self, diagrams: numpy.ndarray) -> numpy.ndarray:
        assert self.fitted_ is True, 'fit() must be called before transform()'
        
        def transform_batch(batch: numpy.ndarray) -> numpy.ndarray:
            batch = self.filtering_.transform(batch)
            return numpy.hstack([
                self.calc_betti_features_            (batch),
                self.calc_landscape_features_        (batch),
                self.calc_silhouette_features_       (batch),
                self.calc_entropy_features_          (batch),
                self.calc_number_of_points_features_ (batch),
                self.calc_heat_features_             (batch),
                self.calc_persistence_image_features_(batch),
                self.calc_lifetime_features_         (batch)
            ])
        
        loop = range(0, len(diagrams), self.batch_size_)
        features = joblib.Parallel(return_as = 'generator', n_jobs = self.n_jobs_)(
            joblib.delayed(transform_batch)(diagrams[batch_start:batch_start + self.batch_size_])
            for batch_start in loop
        )

        collector = cvtda.logging.logger().pbar(features, total = len(loop), desc = 'DiagramVectorizer: batch')
        return numpy.vstack(list(collector))



    def determine_filtering_epsilon_(self, diagrams: numpy.ndarray) -> float:
        life = (diagrams[:, :, 1] - diagrams[:, :, 0]).flatten()
        return numpy.percentile(life[life != 0], self.filtering_percentile_)

    def calc_perdim_sequence_stats_(self, data: numpy.ndarray) -> numpy.ndarray:
        return numpy.hstack([
            cvtda.utils.sequence2features(data[:, dim, :], reduced = self.reduced_)
            for dim in range(data.shape[1])
        ])
    


    def calc_betti_features_(self, diagrams: numpy.ndarray) -> numpy.ndarray:
        betti_curves = self.betti_curve_.transform(diagrams)
        return self.calc_perdim_sequence_stats_(betti_curves)
        
    def calc_landscape_features_(self, diagrams: numpy.ndarray) -> numpy.ndarray:
        n_layers = self.persistence_landscape_.n_layers
        landscape = self.persistence_landscape_.transform(diagrams)
        return numpy.hstack([
            self.calc_perdim_sequence_stats_(landscape[:, layer::n_layers, :])
            for layer in range(n_layers)
        ])

    def calc_silhouette_features_(self, diagrams: numpy.ndarray) -> numpy.ndarray:
        return numpy.hstack([
            self.calc_perdim_sequence_stats_(silhouette.transform(diagrams))
            for silhouette in self.silhouettes_
        ])

    def calc_entropy_features_(self, diagrams: numpy.ndarray) -> numpy.ndarray:
        return self.persistence_entropy_.transform(diagrams)
    
    def calc_number_of_points_features_(self, diagrams: numpy.ndarray) -> numpy.ndarray:
        return self.number_of_points_.transform(diagrams)
    
    def calc_heat_features_(self, diagrams: numpy.ndarray) -> numpy.ndarray:
        flat_shape = (len(diagrams), len(self.homology_dimensions_), -1)
        return numpy.hstack([
            self.calc_perdim_sequence_stats_(heat_kernel.transform(diagrams).reshape(flat_shape))
            for heat_kernel in self.heat_kernels_
        ])
        
    def calc_persistence_image_features_(self, diagrams: numpy.ndarray) -> numpy.ndarray:
        flat_shape = (len(diagrams), len(self.homology_dimensions_), -1)
        return numpy.hstack([
            self.calc_perdim_sequence_stats_(persistence_image.transform(diagrams).reshape(flat_shape))
            for persistence_image in self.persistence_images_
        ])

    def calc_lifetime_features_(self, diagrams: numpy.ndarray) -> numpy.ndarray:
        birth, death, dim = diagrams[:, :, 0], diagrams[:, :, 1], diagrams[:, :, 2]
        bd2 = (birth + death) / 2.0
        life = death - birth

        bd2_bulk = [ ]
        life_bulk = [ ]
        for d in self.homology_dimensions_:
            mask = (dim != d) | (life < self.filtering_epsilon_)
            bd2_bulk.append(numpy.ma.array(bd2, mask = mask))
            life_bulk.append(numpy.ma.array(life, mask = mask))

        return numpy.hstack([
            self.calc_perdim_sequence_stats_(numpy.ma.stack(bd2_bulk, axis = 1)),
            self.calc_perdim_sequence_stats_(numpy.ma.stack(life_bulk, axis = 1))
        ])



import typing

import numpy
import sklearn.metrics
import matplotlib.pyplot as plt

def estimate_quality(
    y_pred_proba: numpy.ndarray,
    y_true: numpy.ndarray,
    ax: typing.Optional[plt.Axes] = None,
    label_names: typing.Optional[typing.List[str]] = None,
    confusion_matrix_include_values: bool = True
) -> dict:
    if label_names is None:
        label_names = list(range(y_pred_proba.shape[1]))

    y_pred = numpy.argmax(y_pred_proba, axis = 1)
    if ax:
        sklearn.metrics.ConfusionMatrixDisplay.from_predictions(
            y_true, y_pred, ax = ax, colorbar = False,
            display_labels = label_names, include_values = confusion_matrix_include_values
        )
        ax.set_xticks(ax.get_xticks(), labels = ax.get_xticklabels(), rotation = 45, ha = "right", rotation_mode = "anchor")
        ax.set_xlabel(None)
        ax.set_ylabel(None)

    numeric_labels = list(range(y_pred_proba.shape[1]))
    auc_roc_input = y_pred_proba
    multi_class = 'ovo'
    average = 'macro'
    
    if y_pred_proba.shape[1] == 2:
        auc_roc_input = y_pred_proba[:, 1]
        multi_class = 'raise'
        average = 'binary'
    
    metrics = {
        'Accuracy':       sklearn.metrics.accuracy_score      (y_true, y_pred),
        'AUC-ROC':        sklearn.metrics.roc_auc_score       (y_true, auc_roc_input, multi_class = multi_class, labels = numeric_labels),
        'Precision':      sklearn.metrics.precision_score     (y_true, y_pred,        average = average,   zero_division = 0),
        'Recall':         sklearn.metrics.recall_score        (y_true, y_pred,        average = average),
        'F1-score':       sklearn.metrics.f1_score            (y_true, y_pred,        average = average)
    }
    for k in [ 2, 3, 5, 7, 9, 15, 30 ]:
        if y_pred_proba.shape[1] > k:
            metrics[f'TOP-{k} Accuracy'] = sklearn.metrics.top_k_accuracy_score(y_true, y_pred_proba, k = k, labels = numeric_labels)
    return metrics



from .DiagramVectorizer import DiagramVectorizer
from .FeatureExtractor import FeatureExtractor
from .FiltrationsExtractor import FiltrationExtractor
from .FiltrationsExtractor import FiltrationsExtractor
from .GeometryExtractor import GeometryExtractor
from .GreyscaleExtractor import GreyscaleExtractor
from .PointCloudsExtractor import PointCloudsExtractor



import typing

import numpy


def deduce_depth(data):
    if isinstance(data, numpy.ndarray) or isinstance(data, list):
        return deduce_depth(data[0]) + 1
    return 0

def hstack(data: typing.List[numpy.ndarray], force_numpy: bool):
    if force_numpy:
        return numpy.hstack(data)
    else:
        if len(data) == 0: return []
        data = [ item for item in data if len(item) > 0 ]
        assert len(set([ len(item) for item in data ])) == 1

        for i in range(len(data)):
            if deduce_depth(data[i]) == 3:
                data[i] = numpy.expand_dims(data[i], 1)

            if deduce_depth(data[i]) != 4:
                print(type(data[i]))
                print(len(data[i]))
                print(data[i].shape)

            assert deduce_depth(data[i]) == 4, f"Bad depth: {deduce_depth(data[i])}"

        return [ sum([ list(item[i]) for item in data ], []) for i in range(len(data[0])) ]



import numpy
import sklearn.base

def process_iter(transformer: sklearn.base.TransformerMixin, data: numpy.ndarray, do_fit: bool, *args, **kwargs):
    if do_fit:
        transformer.fit(data, *args, **kwargs)
    return transformer.transform(data, *args, **kwargs)



from .duplicates import DuplicateFeaturesRemover
from .image2pointcloud import image2pointcloud
from .rgb2gray import rgb2gray
from .rgb2hsv import rgb2hsv
from .sequence2features import sequence2features
from .set_random_seed import set_random_seed
from .spread_points import spread_points



import typing

import numpy
import gtda.homology

import cvtda.utils
import cvtda.logging

from . import utils
from .interface import TopologicalExtractor


class PointCloudsExtractor(TopologicalExtractor):
    def __init__(
        self,
        n_jobs: int = -1,
        reduced: bool = True,
        only_get_from_dump: bool = False,
        return_diagrams: bool = False,
        **kwargs
    ):
        super().__init__(
            supports_rgb = True,
            n_jobs = n_jobs,
            reduced = reduced,
            only_get_from_dump = only_get_from_dump,
            return_diagrams = return_diagrams,
            **kwargs
        )

        self.persistence_ = gtda.homology.VietorisRipsPersistence(homology_dimensions = [0, 1, 2], n_jobs = self.n_jobs_)

    def get_diagrams_(self, images: numpy.ndarray, do_fit: bool, dump_name: typing.Optional[str] = None):
        cvtda.logging.logger().print(f"PointCloudsExtractor: processing {dump_name}, do_fit = {do_fit}")
        
        point_clouds = cvtda.utils.image2pointcloud(images, self.n_jobs_)
        return utils.process_iter_dump(self.persistence_, point_clouds, do_fit, self.diagrams_dump_(dump_name))



import abc
import typing

T = typing.TypeVar("T")

class BaseDumper(abc.ABC, typing.Generic[T]):
    current_dumper = None
    
    def __enter__(self):
        self.__previous = BaseDumper.current_dumper
        BaseDumper.current_dumper = self

    def __exit__(self, *args):
        BaseDumper.current_dumper = self.__previous

    @abc.abstractmethod
    def execute(self, function: typing.Callable[[typing.Any], T], name: str, *function_args) -> T:
        pass

    @abc.abstractmethod
    def save_dump(self, data: T, name: str):
        pass

    @abc.abstractmethod
    def has_dump(self, name: str) -> bool:
        pass

    def get_dump(self, name: str) -> T:
        assert self.has_dump(name), f"There is no dump at {name}"
        return self.get_dump_impl_(name)
    
    @abc.abstractmethod
    def get_dump_impl_(self, name: str) -> T:
        pass



import typing

import numpy
import torch
import joblib
import torch.utils.data
import torchph.nn.slayer

import cvtda.logging


def transform(diagram: torch.Tensor, dim: int):
    dim_filter = (diagram[:, 2] == dim)
    non_degenerate_filter = (diagram[:, 0] < diagram[:, 1])
    rotation = torchph.nn.slayer.UpperDiagonalThresholdedLogTransform(0.01)
    return rotation(diagram[dim_filter & non_degenerate_filter][:, 0:2])

def process_diagram(diags: torch.Tensor):
    diagrams, non_dummy_points = [], []
    for dim in diags[:, :, 2].unique(sorted = False):
        diags_dim = [ transform(diag, dim) for diag in diags ]
        processed = torchph.nn.slayer.prepare_batch(diags_dim)
        diagrams.append(processed[0].cpu())
        non_dummy_points.append(processed[1].cpu())
    return diagrams, non_dummy_points


class Dataset(torch.utils.data.Dataset):
    def __init__(
        self,
        images: numpy.ndarray,
        diagrams: typing.Optional[typing.List[numpy.ndarray]], # n_items x n_diagrams x n_points x 3
        features: numpy.ndarray,
        labels: numpy.ndarray,

        n_jobs: int = -1,
        device: torch.device = torch.device('cuda'),
    ):
        self.n_jobs_ = n_jobs
        self.device_ = device

        if images is not None:
            self.images = torch.tensor(images, dtype = torch.float32)
            if len(self.images.shape) == 4:
                self.images = self.images.permute((0, 3, 1, 2))
            else:
                assert len(self.images.shape) == 3
                self.images = self.images.unsqueeze(1)
            
        self.labels = torch.tensor(labels, dtype = torch.long)
        self.features = torch.tensor(features, dtype = torch.float32)
        self.raw_diagrams = diagrams
        
        if diagrams is None:
            return

        diagrams = [
            torch.tensor(numpy.array([ item[num_diagram] for item in diagrams ]), dtype = torch.float32)
            for num_diagram in range(len(diagrams[0]))
        ]
        diagrams = joblib.Parallel(n_jobs = self.n_jobs_)(
            joblib.delayed(process_diagram)(d)
            for d in cvtda.logging.logger().pbar(diagrams, desc = "Dataset: processing diagrams")
        )

        self.diagrams, self.non_dummy_points = [], []
        for diag, ndp in diagrams:
            self.diagrams.extend(diag)
            self.non_dummy_points.extend(ndp)

        cvtda.logging.logger().print(
            f"Constructed a dataset of {len(self.images)} images of shape {self.images[0].shape} " +
            f"with {len(self.diagrams)} diagrams and {self.features.shape[1]} features"
        )

    def __len__(self):
        return len(self.images)

    def get_labels(self, idxs, device: typing.Optional[torch.device] = None):
        device = device if device is not None else self.device_
        return self.labels[idxs].to(device)
    
    def get_label(self, idx, device: typing.Optional[torch.device] = None):
        device = device if device is not None else self.device_
        return self.labels[idx].to(device)

    def get_diagrams(self, idxs, device: typing.Optional[torch.device] = None):
        device = device if device is not None else self.device_
        output = [ ]
        for diag, ndp in zip(self.diagrams, self.non_dummy_points):
            output.append(diag[idxs].to(device))
            output.append(ndp[idxs].to(device))
        return output

    def get_diagram(self, idx, device: typing.Optional[torch.device] = None):
        device = device if device is not None else self.device_
        output = [ ]
        for diag, ndp in zip(self.diagrams, self.non_dummy_points):
            output.append(diag[idx].to(device))
            output.append(ndp[idx].to(device))
        return output



import abc
import typing

T = typing.TypeVar("T")

class BaseLogger(abc.ABC):
    current_logger = None
    
    def __enter__(self):
        self.__previous = BaseLogger.current_logger
        BaseLogger.current_logger = self

    def __exit__(self, *args):
        BaseLogger.current_logger = self.__previous

    @abc.abstractmethod
    def print(self, data: T) -> None:
        pass

    @abc.abstractmethod
    def pbar(
        self,
        data: typing.Iterable[T],
        total: typing.Optional[int] = None,
        desc: typing.Optional[str] = None
    ) -> typing.Iterable[T]:
        pass

    @abc.abstractmethod
    def zip(
        self,
        *iterables, 
        desc: typing.Optional[str] = None
    ):
        pass

    @abc.abstractmethod
    def set_pbar_postfix(self, pbar: typing.Any, data: dict):
        pass



import numpy
import joblib
import typing

import cvtda.logging


def image2pointcloud(images: numpy.ndarray, n_jobs: int = -1) -> typing.List[numpy.ndarray]:
    def _impl(image: numpy.ndarray) -> numpy.ndarray:
        width, height = image.shape[0:2]
        x = numpy.indices((width, height))[0]
        y = numpy.indices((width, height))[1]
        return numpy.dstack([ x, y, image ]).reshape((width * height, -1))

    return list(
        joblib.Parallel(n_jobs = n_jobs)(
            joblib.delayed(_impl)(img)
            for img in cvtda.logging.logger().pbar(images, desc = "image2pointcloud")
        )
    )
